/* $Id: select.c,v 1.1 1993/10/25 22:35:50 cwikla Exp $ */

/*
**
** Copyright (c) 1993 by Wolfram Research, Inc.  All Rights Reserved.
** 
** WOLFRAM RESEARCH, INC. ("WRI") DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN
** THIS FILE IS FREE FROM BUGS AND OMISSIONS; THE SOFTWARE IS PROVIDED AS IS.
** WRI MAKES NO REPRESENTATIONS, EXPRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE,
** INCLUDING WITHOUT LIMITATIONS ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR
** FITNESS FOR A PARTICULAR PURPOSE, ALL OF WHICH IS EXPRESSLY DISCLAIMED.  WRI
** DOES NOT WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS OR THAT ITS
** OPERATION WILL BE UNINTERRUPTED OR ERROR FREE.
** 
** WRI DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN THIS FILE DOES NOT
** INFRINGE UPON THE INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTIES.
** 
** USE OF THE SOFTWARE CONTAINED IN THIS FILE IS ENTIRELY AT YOUR OWN RISK.
** 
** To the extent that it is authorized to do so, Wolfram Research grants the
** recipient of this software permission to use the software for his/her own
** personal purposes subject the user's acceptance of the terms stated above
** and the user's inclusion of Wolfram Reseach's copyright notice and disclaimers
** in any reproduction and/or derivative of this file.  
** 
**
**
**	John L. Cwikla
**	cwikla@wri.com
** 	Wolfram Research, Inc
*/


/*
** Generic selection mechanism.	Should probably not allow cut-buffers or XA_SECONDARY.	
*/

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Atoms.h>

#if DEBUGGING
#include <stdio.h>
#endif


static void selectPut(Widget _w, XEvent *_event, String *_argv, Cardinal *_argc);
static void selectGet(Widget _w, XEvent *_event, String *_argv, Cardinal *_argc);
static Boolean convertSelectionCB(Widget _w, Atom *_selection, Atom *_target, Atom *_type, XtPointer *_value,
	unsigned long *_length, int *_format);
static void loseSelectionCB(Widget _w, Atom *_selection);
static void putToSelection(Widget _w, Atom _selection, Atom _type);
static void getFromSelection(Widget _w, Atom _selection, Atom _type);
static Atom StringToAtom(Widget _w, String _string);
static void selectionCB(Widget _w, XtPointer _data, Atom *_selection, Atom *_type, XtPointer _value,
	unsigned long *_length, int *_format);

static Window OwnTheSelection = (Window)NULL; 
static Atom XA_WRI_XWD_RLED_ATOM = 0;

XtActionsRec selectActions[] =
{
	{"selection-put", selectPut},
	{"selection-get", selectGet},
};

void AddSelectionActions(XtAppContext _theAppContext)
{
	XtAppAddActions(_theAppContext, selectActions, XtNumber(selectActions));
}

static Atom StringToAtom(Widget _w, String _string)
{
	Atom atom;
	XrmValue from, to;

#if DEBUGGING
printf("StringToAtom... %s\n", _string);
#endif

	from.size = strlen(_string);
	from.addr = (XtPointer)_string;
	to.size = sizeof(Atom);
	to.addr = (XtPointer)&atom;
	XtConvertAndStore(_w, XtRString, &from, XtRAtom, &to);

#if DEBUGGING
printf("%d atom\n", atom);
#endif

	return atom;
} 

static void selectGet(Widget _w, XEvent *_event, String *_argv, Cardinal *_argc)
{
	Atom selectAtom, typeAtom;
	Cardinal numParms = 0;
	XrmValue from, to;

	selectAtom = XA_PRIMARY;
	typeAtom = StringToAtom(_w, "_WRI_XWD_RLED"); 

	if (*_argc > 1)
	{
		XtAppWarningMsg(XtWidgetToApplicationContext(_w),
			"selection-get", "arguments", "TooManyArgs",
			"selection-get can take up to one (1) argument", (String *)NULL, &numParms);
	}
	else
	if (*_argc == 1)
		selectAtom = StringToAtom(_w, *_argv);

	getFromSelection(_w, selectAtom, typeAtom);
}

static void selectPut(Widget _w, XEvent *_event, String *_argv, Cardinal *_argc)
{
	Atom selectAtom, typeAtom;
	Cardinal numParms = 0;
	XrmValue from, to;

	selectAtom = XA_PRIMARY;
	typeAtom = StringToAtom(_w, "_WRI_XWD_RLED");

	if (*_argc > 1)
	{
		XtAppWarningMsg(XtWidgetToApplicationContext(_w), 
			"selection-put", "arguments", "TooManyArgs",
			"selection-put can take up to one (1) argument", (String *)NULL, &numParms);
	}
	else
	if (*_argc == 1)
		selectAtom = StringToAtom(_w, *_argv);

#if DEBUGGING
printf("selectionput\n");
#endif

	putToSelection(_w, selectAtom, typeAtom);
}


static Boolean convertSelectionCB(Widget _w, Atom *_selection, Atom *_target, Atom *_type, XtPointer *_value,
	unsigned long *_length, int *_format)
{
	unsigned int kinds;
	int selSize;
	char *selection;
	XSelectionRequestEvent* req = XtGetSelectionRequest(_w, *_selection, (XtRequestId)NULL);

/*
** ICCCM Compliancy...
*/

#if DEBUGGING
printf("Convert selection\n");
#endif

	if (*_target == StringToAtom(_w, "TARGETS")) /* XA_TARGETS(XtDisplay(_w)))	*/
	{
		Atom *targetP;
		Atom *tP;

#if DEBUGGING
printf("XA_TARGETS\n");
#endif
		tP = targetP = (Atom *)XtMalloc(sizeof(Atom)*2);
		*targetP++ = StringToAtom(_w, "_WRI_XWD_RLED");
		*targetP++ = XA_LENGTH(XtDisplay(_w));
		*_value = tP;
		*_length = 2;
		*_type = XA_ATOM;
		*_format = 32;
		return TRUE;
	}

	if (*_target == StringToAtom(_w, "_WRI_XWD_RLED"))
	{
#if DEBUGGING
printf("_WRI_XWD_RLED\n");
fflush(stdout);
#endif
			selection = (char *)GetImageDataToPut(&selSize);
			*_type = StringToAtom(_w, "_WRI_XWD_RLED");
			*_value = (XtPointer)selection;
			*_length = selSize;
			*_format = 8;
			return TRUE;
	}
	else
	if (*_target == XA_LENGTH(XtDisplay(_w)))
	{
#if DEBUGGING
printf("XA_LENGTH\n");
#endif
			selection = (char *)GetImageDataToPut(&selSize);
			*_type = XA_INTEGER;
			*_value = (XtPointer)selSize;
			*_length = 1L;
			*_format = 32;
			return TRUE;
	}

#if DEBUGGING
printf("Unknown target\n");
printf("%s \n", XmuGetAtomName(XtDisplay(_w), *_target));
#endif

	return FALSE;
}

static void loseSelectionCB(Widget _w, Atom *_selection)
{
	Arg warg;
	OwnTheSelection = (Window)NULL; 
}

static void putToSelection(Widget _w, Atom _selection, Atom _type)
{
	if (XtOwnSelection(_w, _selection, XtLastTimestampProcessed(XtDisplay(_w)), convertSelectionCB, loseSelectionCB, NULL))
	{
	  OwnTheSelection = XtWindow(_w);
      MarkCurrentGetBuffer();
	}

#if DEBUGGING
printf("%d OwnTheSelection\n", OwnTheSelection);
#endif
}

static void getFromSelection(Widget _w, Atom _selection, Atom _type)
{
	XtAppContext app;

	if (OwnTheSelection == XtWindow(_w))
		return;
	else
		XtGetSelectionValue(_w, _selection, _type, selectionCB, NULL, XtLastTimestampProcessed(XtDisplay(_w)));

#if DEBUGGING
printf("getFromSelection..\n");
#endif
}

static void selectionCB(Widget _w, XtPointer _data, Atom *_selection, Atom *_type, XtPointer _value,
	unsigned long *_length, int *_format)
{
	if (*_length == 0)
		return;

	if (*_type == StringToAtom(_w, "_WRI_XWD_RLED"))
	{
		if (*_format == 8)
		{
#if DEBUGGING
printf("selectionCB %d\n", *_length);
#endif
			PutImageForData(_w, _value, *_length);
		}
#if DEBUGGING
		else
			printf("Wrong format...\n");
#endif
	}
	else
		XtFree((char *)_value);
}


