/* $Id: wd.c,v 1.1 1993/10/25 22:35:53 cwikla Exp $ */

/* $XConsortium: xwud.c,v 1.48 91/07/25 18:02:05 rws Exp $ */
/* Copyright 1985, 1986, 1988 Massachusetts Institute of Technology */

/*
Permission to use, copy, modify, distribute, and sell this software and its
documentation for any purpose is hereby granted without fee, provided that
the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of M.I.T. not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.  M.I.T. makes no representations about the
suitability of this software for any purpose.  It is provided "as is"
without express or implied warranty.
*/

/* xwud - marginally useful raster image undumper */

/*
**
** Copyright (c) 1993 by Wolfram Research, Inc.  All Rights Reserved.
**
** WOLFRAM RESEARCH, INC. ("WRI") DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN
** THIS FILE IS FREE FROM BUGS AND OMISSIONS; THE SOFTWARE IS PROVIDED AS IS.
** WRI MAKES NO REPRESENTATIONS, EXPRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE,
** INCLUDING WITHOUT LIMITATIONS ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR
** FITNESS FOR A PARTICULAR PURPOSE, ALL OF WHICH IS EXPRESSLY DISCLAIMED.  WRI
** DOES NOT WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS OR THAT ITS
** OPERATION WILL BE UNINTERRUPTED OR ERROR FREE.
**
** WRI DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN THIS FILE DOES NOT
** INFRINGE UPON THE INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTIES.
**
** USE OF THE SOFTWARE CONTAINED IN THIS FILE IS ENTIRELY AT YOUR OWN RISK.
**
** To the extent that it is authorized to do so, Wolfram Research grants the
** recipient of this software permission to use the software for his/her own
** personal purposes subject the user's acceptance of the terms stated above
** and the user's inclusion of Wolfram Reseach's copyright notice and disclaimers
** in any reproduction and/or derivative of this file.
**
**
**
**  John L. Cwikla
**  cwikla@wri.com
**  Wolfram Research, Inc
*/

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/XWDFile.h>

#include "wd.h"
#include "misclib.h"

static int getXColors(Display *theDisplay, Visual *theVisual, Colormap theColormap, XColor **colors);

#define lowbit(x) ((x) & (~(x) + 1))

#ifndef TRUE
#define TRUE 1
#endif

extern int MyAllocColor();

XImage *GetXWDImage(Display *theDisplay, Visual *theVisual, Colormap theColormap, int theDepth, unsigned char *data)
{
	unsigned long swaptest = 1;
	XWDFileHeader header;
	Display fakeDpy;
	ScreenFormat fakeFmt;
	XImage *inImage, *outImage;
	XWDColor *colors;
	XColor theColor, *xcolors;
	int x, y, i;
	int numberOfColors;
	char *buffer;
	int bufferSize;
	unsigned char *ptr;
	ColorInfo ci;

	ptr = data;

	bcopy(ptr, &header, sizeof(header));

	if (*(char *) &swaptest)
		SwapLong((char *)&header, sizeof(header));

	ptr += sizeof(header);

	if ((header.file_version != XWD_FILE_VERSION) || (header.header_size < sizeof(header)))
	{
#if DEBUGGING
printf("wrong file version %d instead of %d\n", header.file_version, 
	XWD_FILE_VERSION);
printf("Header size is %d it should be %d\n", header.header_size,
	sizeof(header));
#endif
		return (XImage *)NULL;
	}

	ptr += (header.header_size - sizeof(header));

	fakeDpy.byte_order = (int) header.byte_order;
	fakeDpy.bitmap_unit = (int) header.bitmap_unit;
	fakeDpy.bitmap_bit_order = (int) header.bitmap_bit_order;
	fakeDpy.pixmap_format = &fakeFmt;
	fakeDpy.nformats = 1;
	fakeFmt.depth = (int) header.pixmap_depth;
	fakeFmt.bits_per_pixel = (int) header.bits_per_pixel;

	inImage = XCreateImage(&fakeDpy, NULL,
				(int) header.pixmap_depth,
				(int) header.pixmap_format,
				(int) header.xoffset, NULL,
				(int) header.pixmap_width/*-1*/,
				(int) header.pixmap_height/*-1*/,
				(int) header.bitmap_pad,
				(int) header.bytes_per_line);

	inImage->red_mask = header.red_mask;
	inImage->green_mask = header.green_mask;
	inImage->blue_mask = header.blue_mask;

	/* read in the color map buffer */
	numberOfColors = header.ncolors;

	colors = (XWDColor *)XtMalloc((unsigned) numberOfColors * sizeof(XWDColor));

	if (!colors)
	{
		XDestroyImage(inImage);
		return (XImage *)NULL;
	}

	xcolors = (XColor *)XtMalloc((unsigned) numberOfColors * sizeof(XColor));

	if (!xcolors)
	{
		XtFree(colors);
		XDestroyImage(inImage);
		return (XImage *)NULL;
	}

	bcopy(ptr, colors, sizeof(XWDColor)*numberOfColors);

	ptr += (sizeof(XColor)*numberOfColors);

	if (*(char *) &swaptest) 
	{
		for (i = 0; i < numberOfColors; i++) 
		{
			SwapLong((char *) &colors[i].pixel, sizeof(CARD32));
			SwapShort((char *) &colors[i].red, 3 * sizeof(CARD16));
		}
	}

	for (i=0;i<numberOfColors;i++)
	{
		xcolors[i].pixel = (unsigned long)colors[i].pixel;
		xcolors[i].red = (short)colors[i].red;
		xcolors[i].blue = (short)colors[i].blue;
		xcolors[i].green = (short)colors[i].green;
	}

	XtFree(colors);

	bufferSize = XImageSize(inImage);

	if((buffer = (char *)XtMalloc(bufferSize)) == NULL)
	{
		XDestroyImage(inImage);
		XtFree((char *)xcolors);
		return (XImage *)NULL;
	}

	bcopy(ptr, buffer, sizeof(char)*bufferSize);

#if 0
	if (inImage->depth == 1) 
			inImage->format = XYBitmap;
#endif

	inImage->data = buffer;

	outImage = XCreateImage(theDisplay, theVisual, theDepth, /* (theDepth == 1) ? XYBitmap : */ inImage->format,
		inImage->xoffset, NULL, inImage->width, inImage->height, XBitmapPad(theDisplay), 0);

	outImage->data = (char *)XtMalloc(XImageSize(outImage));

	if (outImage->data == NULL)
	{
		XDestroyImage(inImage);
		XDestroyImage(outImage);
		XtFree((char *)xcolors);
		XtFree((char *)buffer);
		return (XImage *)NULL;
	}

	ci.colormap = theColormap;
	ci.numberOfColors = numberOfColors;
	ci.colors = xcolors;
	ci.allocFunc = MyAllocColor;
	ci.defaultPixel = BlackPixel(theDisplay, DefaultScreen(theDisplay));
	ci.visualClass = header.visual_class;

	if ( (header.visual_class == DirectColor) || (header.visual_class == TrueColor))
	{
		ci.rmask = header.red_mask;
		ci.gmask = header.green_mask;
		ci.bmask = header.blue_mask;
		DoDirectImage(theDisplay, &ci, inImage, outImage, TRUE);
	}
	else
		DoStandardImage(theDisplay, &ci, inImage, outImage, TRUE);

	XDestroyImage(inImage);
	XtFree((char *)xcolors);

	return outImage;
}

unsigned char *PutXWDImage(Display *theDisplay, Visual *theVisual, Colormap theColormap, XImage *image, int *retLen)
{
	unsigned long swaptest = 1;
	XWDFileHeader header;
	int theDepth;
	XColor *xcolors;
	XWDColor *colors;
	int numberOfColors;
	int dataSize;
	char *buffer;
	int bufferSize;
	int i, status;
	unsigned char *data, *ptr;

	theDepth = image->depth;

	*retLen = 0;

	bufferSize = XImageSize(image);

#if (DEBUGGING > 3)
printf("bufferSize %d\n", bufferSize);
#endif

	numberOfColors = getXColors(theDisplay, theVisual, theColormap, &xcolors);

#if (DEBUGGING > 3)
printf("%d numberOfColors\n", numberOfColors);
#endif

	if (!xcolors)
		return (unsigned char *)NULL;

	colors = (XWDColor *)XtMalloc(sizeof(XWDColor) * numberOfColors);

	if (colors == NULL)
	{
		XtFree((char *)xcolors);
		return (unsigned char *)NULL;
	}

	for(i=0;i<numberOfColors;i++)
	{
		colors[i].pixel = (CARD32)xcolors[i].pixel;
		colors[i].red = (CARD16)xcolors[i].red;
		colors[i].blue = (CARD16)xcolors[i].blue;
		colors[i].green = (CARD16)xcolors[i].green;
	}

	XtFree((char *)xcolors);

	dataSize = sizeof(header) + (sizeof(XWDColor) * numberOfColors) + bufferSize;

	data = (unsigned char *)XtMalloc(dataSize);

	if (data == NULL)
	{
		XtFree((char *)colors);
		return (unsigned char *)NULL; 
	}

	ptr = data;

	header.header_size = (CARD32)sizeof(header);
	header.file_version = (CARD32) XWD_FILE_VERSION;
	header.pixmap_format = (CARD32) image->format;
	header.pixmap_depth = (CARD32) image->depth;
	header.pixmap_width = (CARD32) image->width;
	header.pixmap_height = (CARD32) image->height;
	header.xoffset = (CARD32) image->xoffset;
	header.byte_order = (CARD32) image->byte_order;
	header.bitmap_unit = (CARD32) image->bitmap_unit;
	header.bitmap_bit_order = (CARD32) image->bitmap_bit_order;
	header.bitmap_pad = (CARD32) image->bitmap_pad;
	header.bits_per_pixel = (CARD32) image->bits_per_pixel;
	header.bytes_per_line = (CARD32) image->bytes_per_line;
	header.visual_class = (CARD32) theVisual->class;
	header.red_mask = (CARD32) theVisual->red_mask;
	header.green_mask = (CARD32) theVisual->green_mask;
	header.blue_mask = (CARD32) theVisual->blue_mask;
	header.bits_per_rgb = (CARD32) theVisual->bits_per_rgb;
	header.colormap_entries = (CARD32) theVisual->map_entries;
	header.ncolors = (CARD32)numberOfColors;
	header.window_width = (CARD32) image->width;
	header.window_height = (CARD32) image->height;
	header.window_x = (CARD32)0;
	header.window_y = (CARD32)0;
	header.window_bdrwidth = (CARD32)0;

	if (*(char *) &swaptest)
		SwapLong((char *)&header, sizeof(header));

	if (*(char *) &swaptest)
	{
		for (i = 0; i < numberOfColors; i++)
		{
			SwapLong((char *) &colors[i].pixel, sizeof(CARD32));
			SwapShort((char *) &colors[i].red, 3 * sizeof(CARD16));
		}
	}

	bcopy(&header, ptr, sizeof(header)); ptr += sizeof(header);
	bcopy(colors, ptr, sizeof(XColor)*numberOfColors); ptr += (sizeof(XColor)*numberOfColors);
	bcopy(image->data, ptr, sizeof(char)*bufferSize);

	XtFree((char *)colors);

	*retLen = dataSize;

#if DEBUGGING
printf("%d retLen\n", *retLen);
#endif


	return data;
}

static int getXColors(Display *theDisplay, Visual *theVisual, Colormap theColormap, XColor **colors)
{
	int i, ncolors;
	Colormap cmap;
	Visual *visual;
	Display *dpy;

	dpy = theDisplay;
	visual = theVisual;
	cmap = theColormap;

	ncolors = visual->map_entries;
	*colors = (XColor *)XtMalloc(sizeof(XColor)*ncolors);
 
	if ((*colors) == (XColor *)NULL)
		return 0;

	if (visual->class == DirectColor || visual->class == TrueColor) 
	{
		unsigned long red, green, blue, red1, green1, blue1;

		red = green = blue = 0;
		red1 = lowbit(visual->red_mask);
		green1 = lowbit(visual->green_mask);
		blue1 = lowbit(visual->blue_mask);
		for (i=0; i<ncolors; i++) 
		{
			(*colors)[i].pixel = red|green|blue;
			(*colors)[i].pad = 0;
			red += red1;
			if (red > visual->red_mask)
				red = 0;
			green += green1;
			if (green > visual->green_mask)
				green = 0;
			blue += blue1;
			if (blue > visual->blue_mask)
				blue = 0;
		}
	} 
	else 
	{
		for (i=0; i<ncolors; i++) 
		{
			(*colors)[i].pixel = i;
			(*colors)[i].pad = 0;
		}
	}

	XQueryColors(dpy, cmap, *colors, ncolors);

	return(ncolors);
}

unsigned int XImageSize(XImage *image)
{
	if (image->format != ZPixmap)
		return(image->bytes_per_line * image->height * image->depth);

	return((unsigned)image->bytes_per_line * image->height);
}
