/* $Id: xwdclip.c,v 1.2 1993/10/25 23:38:37 cwikla Exp $ */

/*
**
** Copyright (c) 1993 by Wolfram Research, Inc.  All Rights Reserved.
** 
** WOLFRAM RESEARCH, INC. ("WRI") DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN
** THIS FILE IS FREE FROM BUGS AND OMISSIONS; THE SOFTWARE IS PROVIDED AS IS.
** WRI MAKES NO REPRESENTATIONS, EXPRESS OR IMPLIED, WITH RESPECT TO THE SOFTWARE,
** INCLUDING WITHOUT LIMITATIONS ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR
** FITNESS FOR A PARTICULAR PURPOSE, ALL OF WHICH IS EXPRESSLY DISCLAIMED.  WRI
** DOES NOT WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS OR THAT ITS
** OPERATION WILL BE UNINTERRUPTED OR ERROR FREE.
** 
** WRI DOES NOT WARRANT THAT THE SOFTWARE CONTAINED IN THIS FILE DOES NOT
** INFRINGE UPON THE INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTIES.
** 
** USE OF THE SOFTWARE CONTAINED IN THIS FILE IS ENTIRELY AT YOUR OWN RISK.
** 
** To the extent that it is authorized to do so, Wolfram Research grants the
** recipient of this software permission to use the software for his/her own
** personal purposes subject the user's acceptance of the terms stated above
** and the user's inclusion of Wolfram Reseach's copyright notice and disclaimers
** in any reproduction and/or derivative of this file.  
** 
**
**
**	John L. Cwikla
**	cwikla@wri.com
** 	Wolfram Research, Inc
*/


#include <X11/Xmd.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include <X11/Shell.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/ScrolledW.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/Form.h>

#include <stdio.h>

#include "wd.h"
#include "rled.h"

#include "xwdclip.h"
#include "getfile.h"

static CutBuffer TheBuffers[MAX_BUFFER];
static int TheFrontBuffer = 0;
static int GetBuffer;
static CutBuffer *TheCurrentBuffer;

static Widget DA;
static Widget SW;
static Widget TopLevel;
static Widget TheLabel;

static Cursor TheWaitCursor;

static GC theGC; 
static Display *TheDisplay = NULL;
static Drawable TheWindow;
static Colormap TheColormap;
static Visual *TheVisual;
static int TheDepth;

#define NAME "xwdclip"
#define CLASSNAME "Xwdclip"

#ifndef MIN
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif

typedef struct _ButtonStruct
{
	char *name;
	int value;
} ButtonStruct;

static ButtonStruct TheButtons[] =
{
	{"Next", NEXT},
	{"Prev", PREV},
	{"Clear", CLEAR},
	{"Save...", SAVE},
	{"Open...", OPEN},
	{"Quit", QUIT},
};

static XmString TheLabelStrings[MAX_BUFFER];

int MyAllocColor(Display *dpy, Colormap cmap, XColor *color)
{
	return XAllocColor(dpy, cmap, color);
}

void EmptyBuffer(int _bufferNumber)
{
	CutBuffer *buf;

	if (_bufferNumber >= MAX_BUFFER)
		return;

	buf = TheBuffers+_bufferNumber;

	if (buf->image)
		XDestroyImage(buf->image);
	if (buf->data)
		XtFree(buf->data);

	buf->image = (XImage *)NULL;
	buf->data = (unsigned char *)NULL;
	buf->length = 0;
}

void ClearAllBuffers()
{
	int i;

	for(i=0;i<MAX_BUFFER;i++)
		EmptyBuffer(i);
}

void MarkCurrentGetBuffer()
{
	GetBuffer = TheFrontBuffer;
}

void SetCurrentBuffer(int _bufferNumber)
{
	Arg warg;

	if (_bufferNumber >= MAX_BUFFER)
		_bufferNumber = 0;
	else
	if (_bufferNumber < 0)
		_bufferNumber = (MAX_BUFFER-1);

	TheCurrentBuffer = TheBuffers+_bufferNumber;
	TheFrontBuffer = _bufferNumber;

	XtSetArg(warg, XmNlabelString, TheLabelStrings[TheFrontBuffer]); 
	XtSetValues(TheLabel, &warg, 1);

	XFlush(TheDisplay);
}

void NextBuffer()
{
	SetCurrentBuffer(TheFrontBuffer+1);
}

void PrevBuffer()
{
	SetCurrentBuffer(TheFrontBuffer-1);
}

static void init()
{
	int i;
	char buf[3];

	for(i=0;i<MAX_BUFFER;i++)
	{
		TheBuffers[i].image = (XImage *)NULL;
		TheBuffers[i].data = (unsigned char *)NULL;
		TheBuffers[i].length = 0;

		sprintf(buf, "%d", i);
		TheLabelStrings[i] = XmStringCreateLtoR(buf, XmSTRING_DEFAULT_CHARSET);
	}

	TheWaitCursor = XCreateFontCursor(TheDisplay, XC_watch);

	SetCurrentBuffer(0);
}

void SetCursorWait(Boolean _state)
{
	if (_state)
		XDefineCursor(XtDisplay(TopLevel), XtWindow(TopLevel), TheWaitCursor);
	else
		XUndefineCursor(XtDisplay(TopLevel), XtWindow(TopLevel));

	XFlush(XtDisplay(TopLevel));
}

static void expose()
{
	Arg warg[2];
	int n;

	n = 0;

	XClearWindow(TheDisplay, TheWindow);
	if (TheCurrentBuffer->image)
	{
		XPutImage(TheDisplay, TheWindow, theGC, TheCurrentBuffer->image, 0, 0, 0, 0,
			TheCurrentBuffer->image->width,
			TheCurrentBuffer->image->height);

		XtSetArg(warg[n], XtNwidth, TheCurrentBuffer->image->width); n++;
		XtSetArg(warg[n], XtNheight, TheCurrentBuffer->image->height); n++;
	}
	else
	{
		XtSetArg(warg[n], XtNwidth, 50); n++;
		XtSetArg(warg[n], XtNheight, 50); n++;
	}

	XtSetValues(DA, warg, n);
} 

static void exposeCB(Widget _w, XtPointer call, XmDrawingAreaCallbackStruct *dacs)
{
	if (dacs->event->xexpose.count != 0)
		return;

	expose();
}

static void buttonCB(Widget _w, XtPointer call, XtPointer nothing)
{
	int value;

	value = (int)call;

	SetCursorWait(TRUE);

	switch(value)
	{
		case QUIT:
			exit(0);
		case SAVE:
		{
			INT32 realLength;
			unsigned char *realData;
			FILE *fid;
			char *filename;

			if (TheCurrentBuffer->data && TheCurrentBuffer->length)
				realData = RunLengthDecode(TheCurrentBuffer->data, TheCurrentBuffer->length, &realLength);

			if (realData == NULL)
				break;

			filename = GetAFileName(DA);

			if (filename == NULL)
				break;

			if ((fid = fopen(filename, "w")) != (FILE *)NULL)
			{
				fwrite((char *)realData, sizeof(char), realLength, fid);
				fclose(fid);
			}

			XtFree(filename);

			free(realData);
			break;
		}
		case CLEAR:
			EmptyBuffer(TheFrontBuffer);
			expose();
			break;
		case NEXT:
			NextBuffer();
			expose();
			break;
		case PREV:
			PrevBuffer();
			expose();
			break;
		case OPEN:
		{
			FILE *fid;
			unsigned char *data;
			int size;
			char *filename;

			filename = GetAFileName(DA);

			if (filename == NULL)
				break;

			EmptyBuffer(TheFrontBuffer);

			if ((fid = fopen(filename, "r")) != (FILE *)NULL)
			{
				fseek(fid, 0, 2);
				size = ftell(fid);
				fseek(fid, 0, 0);

				data = (unsigned char *)malloc(sizeof(unsigned char)*size);

				if (data)
				{
					fread((char *)data, sizeof(char), size, fid);
					TheCurrentBuffer->data = RunLengthEncode(data, size, &TheCurrentBuffer->length);

					TheCurrentBuffer->image = GetXWDImage(TheDisplay, TheVisual, TheColormap, TheDepth, data);

					free(data);
				}
			}
			XtFree(filename);
			expose();
		}
		break;
	}

	SetCursorWait(FALSE);
}

void PutImageForData(Widget _w, unsigned char *data, int length)
{
	unsigned char *realData;
	INT32 realLength;
	XImage *image;

	if (data && length)
		realData = RunLengthDecode(data, length, &realLength);

	if (realData && realLength)
	{
		EmptyBuffer(TheFrontBuffer);

		SetCursorWait(TRUE);

		TheCurrentBuffer->image = GetXWDImage(TheDisplay, TheVisual, TheColormap, TheDepth, realData);
		TheCurrentBuffer->data = data;
		TheCurrentBuffer->length = length;

		SetCursorWait(FALSE);

		free(realData);
	}


	expose();
}

unsigned char *GetImageDataToPut(int *retLen)
{
	int sNum;
	XImage *image;
	unsigned char *data;
	int len;
	CutBuffer *tempBuffer;

	*retLen = 0;

	tempBuffer = &TheBuffers[GetBuffer];

	if (tempBuffer->data && tempBuffer->length)
	{
		*retLen = tempBuffer->length;
		return tempBuffer->data;
	}
	else
	{
		*retLen = 0;
		return (unsigned char *)NULL;
	}
}

main(int _argc, char *_argv[])
{
	XtAppContext theAppContext;
	Widget core, row, row2, button, form;
	int n;
	Arg warg[8];
	int i;

	XtToolkitInitialize();
	theAppContext = XtCreateApplicationContext();

	TheDisplay = XtOpenDisplay(theAppContext, NULL, NAME, CLASSNAME, NULL, 0, &_argc, _argv);


	if (!TheDisplay)
	{
		fprintf(stderr, "%s: can't open TheDisplay, exiting...\n", _argv[0]);
		exit(0);
	}

	AddSelectionActions(theAppContext);

	n = 0;
	TopLevel = XtAppCreateShell(NAME, CLASSNAME, applicationShellWidgetClass, TheDisplay, NULL, 0);

	n = 0;
	XtSetArg(warg[n], XmNorientation, XmHORIZONTAL); n++;
	form = XtCreateManagedWidget("_form", xmFormWidgetClass, TopLevel, warg, n);

	n = 0;
	XtSetArg(warg[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
	XtSetArg(warg[n], XmNtopAttachment, XmATTACH_FORM); n++;
	XtSetArg(warg[n], XmNbottomAttachment, XmATTACH_FORM); n++;
	XtSetArg(warg[n], XmNleftAttachment, XmATTACH_FORM); n++;
	row2 = XtCreateManagedWidget("_rowColumn2", xmRowColumnWidgetClass, form, warg, n);

	n = 0;
	XtSetArg(warg[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
	XtSetArg(warg[n], XmNtopAttachment, XmATTACH_FORM); n++;
	XtSetArg(warg[n], XmNbottomAttachment, XmATTACH_FORM); n++;
	XtSetArg(warg[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg(warg[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
	XtSetArg(warg[n], XmNleftWidget, row2); n++;
	SW = XtCreateManagedWidget("_scrolledWindow", xmScrolledWindowWidgetClass, form, warg, n);

	n = 0;
	XtSetArg(warg[n], XtNwidth, 300); n++;
	XtSetArg(warg[n], XtNheight, 300); n++;
	DA = XtCreateManagedWidget("_drawingArea", xmDrawingAreaWidgetClass, SW, warg, n);

	XtAddCallback(DA, XmNexposeCallback, (XtCallbackProc)exposeCB, NULL);

	for(i=0;i<XtNumber(TheButtons);i++)
	{
		button = XtCreateManagedWidget(TheButtons[i].name, xmPushButtonWidgetClass, row2, warg, n);
		XtAddCallback(button, XmNactivateCallback, (XtCallbackProc)buttonCB, (XtPointer)TheButtons[i].value);
	}

	TheLabel = XtCreateManagedWidget("_label", xmLabelWidgetClass, row2, warg, n);

	init();

	XtRealizeWidget(TopLevel);

	TheColormap = DefaultColormap(TheDisplay, DefaultScreen(TheDisplay));
	TheVisual = DefaultVisual(TheDisplay, DefaultScreen(TheDisplay));
	TheDepth = DefaultDepth(TheDisplay, DefaultScreen(TheDisplay));

	TheWindow = XtWindow(DA);

	theGC = XtGetGC(TopLevel, NULL, 0);

	XtAppMainLoop(theAppContext);
}
